"""
This module implements a :class:`asyncactor.Actor` which works on top of
a DistKV client.
"""

import anyio
from ..util import singleton
from asyncactor.abc import Transport, MonitorStream
from asyncactor import Actor  # noqa

__all__ = [
        "ClientActor",
        "ActorState",
        "DetachedState",
        "PartialState",
        "CompleteState",
    ]

class ClientActor(Actor):
    def __init__(self, client, *a, topic, **kw):
        super().__init__(ClientTransport(client, topic), *a, **kw)

class ClientTransport(Transport):
    """
    This class exports the client's direct messaging interface to the
    actor.
    """
    def __init__(self, client, topic):
        self.client = client
        self.topic = topic

    def monitor(self):
        return ClientMonitor(self)

    async def send(self, payload):
        await self.client.msg_send(self.topic, payload)


class ClientMonitor(MonitorStream):
    async def __aenter__(self):
        self._mon1 = self.transport.client.msg_monitor(self.transport.topic)
        self._mon2 = await self._mon1.__aenter__()
        return self

    def __aexit__(self, *tb):
        return self._mon1.__aexit__(*tb)

    def __aiter__(self):
        self._it = self._mon2.__aiter__()
        return self

    async def __anext__(self):
        msg = await self._it.__anext__()
        return msg.data


# The following events are used by Runner etc. to notify running jobs
# about the current connectivity state.
#
class ActorState:
    """abstract base class for states"""

    pass


@singleton
class DetachedState(ActorState):
    """I am detached, my actor group is not visible"""

    pass


@singleton
class PartialState(ActorState):
    """Some but not all members of my actor group are visible"""

    pass


@singleton
class CompleteState(ActorState):
    """All members of my actor group are visible"""

    pass



"""
This module's job is to run code, resp. to keep it running.


"""

import anyio
from weakref import ref
from asyncactor import NodeList
from asyncactor import PingEvent, TagEvent, UntagEvent, AuthPingEvent
from copy import deepcopy
import psutil
import time
from collections.abc import Mapping

from .actor import ClientActor
from .actor import DetachedState, PartialState, CompleteState, ActorState
from .util import NotGiven

from .exceptions import ServerError
from .obj import AttrClientEntry, ClientRoot, ClientEntry

import logging

logger = logging.getLogger(__name__)

QLEN = 10


class NotSelected(RuntimeError):
    """
    This node has not been selected for a very long time. Something is amiss.
    """

    pass


class RunnerEntry(AttrClientEntry):
    """
    An entry representing some hopefully-running code.

    The code will run some time after ``target`` has passed.
    On success, it will run again ``repeat`` seconds later (if >0).
    On error, it will run ``delay`` seconds later (if >0), multiplied by 2**backoff.

    Arguments:
      code (list): pointer to the code that's to be started.
      data (dict): additional data for the code.
      delay (float): time before restarting the job on error
      repeat (float): time before restarting on success
      target (float): time the job should be started at

    The code runs with these additional keywords::

      _entry: this object
      _client: the DistKV client connection
      _info: a queue to send events to the task. A message of ``None``
          signals that the queue was overflowing and no further messages will
          be delivered.

    Messages are defined in :mod:`distkv.actor`.
    """

    ATTRS = "code data delay ok_after repeat backoff target".split()

    delay = 100  # timedelta, before restarting
    repeat = 0
    target = 0
    backoff = 1.4

    code = ()  # what to execute
    data = None
    scope = None  # scope to kill off
    _comment = None  # used for error entries, i.e. mainly Cancel
    _q = None  # send events to the running task. Async tasks only.
    _running = False  # .run is active. Careful with applying updates.
    _task = None

    def __init__(self, *a, **k):
        self.data = {}  # local data

        super().__init__(*a, **k)

    def __repr__(self):
        return "<%s %r:%r>" % (self.__class__.__name__, self.subpath, self.code)

    @property
    def state(self):
        return self.root.state.follow(*self.subpath, create=None)

    async def run(self):
        if self.code is None:
            return  # nothing to do here

        state = self.state
        try:
            self._running = True
            try:
                if state.node is not None:
                    raise RuntimeError("already running on %s", state.node)
                code = self.root.code.follow(*self.code, create=False)
                data = self.data
                if data is None:
                    data = {}
                else:
                    data = deepcopy(data)

                if code.is_async:
                    data["_info"] = self._q = anyio.create_queue(QLEN)
                    if self.root._active is not None:
                        await self._q.put(self.root._active)
                data["_entry"] = self
                data["_client"] = self.root.client
                data["_cfg"] = self.root.client._cfg

                state.started = time.time()
                state.node = state.root.name

                await state.save(wait=True)
                if state.node != state.root.name:
                    raise RuntimeError(
                        "Rudely taken away from us.", state.node, state.root.name
                    )

                logger.debug("Start %r with %r", self._path, self.code)
                async with anyio.create_task_group() as tg:
                    sc = tg.cancel_scope
                    self.scope = sc
                    oka = getattr(self, 'ok_after', -1)
                    if oka > 0:
                        async def is_ok(oka):
                            await anyio.sleep(oka)
                            state.backoff = 0
                            await state.save()
                            await self.root.err.record_working("run", *self._path)

                        await tg.spawn(is_ok, oka)
                    res = code(**data)
                    if code.is_async is not None:
                        res = await res
                    await sc.cancel()
            finally:
                logger.debug("End %r", self._path)
                self.scope = None
                self._q = None
                t = time.time()

        except BaseException as exc:
            c, self._comment = self._comment, None
            async with anyio.move_on_after(2, shield=True):
                await self.root.err.record_error(
                    "run",
                    *self._path,
                    message="Exception",
                    exc=exc,
                    data=self.data,
                    comment=c
                )
            state.backoff += 1
        else:
            state.result = res
            state.backoff = 0
            await self.root.err.record_working("run", *self._path)
        finally:
            async with anyio.fail_after(2, shield=True):
                if state.node == state.root.name:
                    state.node = None
                self._running = False
                state.stopped = t

                if state.backoff > 0:
                    self.retry = t + (self.backoff ** state.backoff) * self.delay
                else:
                    self.retry = None
                async with anyio.move_on_after(2, shield=True):
                    try:
                        await state.save()
                    except anyio.exceptions.ClosedResourceError:
                        pass
                    except ServerError:
                        logger.exception("Could not save")
                    await self.root.trigger_rescan()

    async def send_event(self, evt):
        if self._q is not None:
            if self._q.qsize() < QLEN - 1:
                await self._q.put(evt)
            elif self._q.qsize() == QLEN - 1:
                await self._q.put(None)
                self._q = None

    async def seems_down(self):
        state = self.state
        state.node = None
        if not self._running:
            await state.save(wait=True)

    async def set_value(self, val):
        c = self.code
        await super().set_value(val)

        # Check whether running code needs to be killed off
        if self.scope is not None:
            if val is NotGiven or c != self.code:
                # The code changed.
                self._comment = "Cancel: Code changed"
                await self.scope.cancel()
            elif not getattr(self, "target", None):
                self._comment = "Cancel: target zeroed"
                await self.scope.cancel()

        await self.root.trigger_rescan()

    async def seen_value(self):
        await super().seen_value()
        await self.root.trigger_rescan()

    async def run_at(self, t: float):
        """Next run at this time.
        """
        self.target = t
        if not self._running:
            await self.save()

    def should_start(self):
        """Tell whether this job might want to be started.

        Returns:
          ``False``: No, it's running (or has run and doesn't restart).
          n>0: wait for n seconds before thinking again.
          n<0: should have been started n seconds ago, do something!

        """

        state = self.state
        if self.code is None:
            return False
        if state.node is not None:
            return False
        assert not state.started or state.stopped

        if self.target > state.started:
            return self.target
        elif state.backoff:
            return state.stopped + self.delay * (self.backoff ** state.backoff)
        elif self.repeat:
            return state.stopped + self.repeat
        elif state.started:
            return False
        else:
            return 0

    def __hash__(self):
        return hash(self.subpath)

    def __eq__(self, other):
        other = getattr(other, "subpath", other)
        return self.subpath == other

    def __lt__(self, other):
        other = getattr(other, "subpath", other)
        return self.subpath < other

    @property
    def age(self):
        return time.time() - self.state.started


class RunnerNode:
    """
    Represents all nodes in this runner group.

    This is used for load balancing and such. TODO.
    """

    seen = 0
    load = 999

    def __new__(cls, root, name):
        try:
            self = root._nodes[name]
        except KeyError:
            self = object.__new__(cls)
            self.root = root
            self.name = name
            root._nodes[name] = self
        return self

    def __init__(self, *a, **k):
        pass


class StateEntry(AttrClientEntry):
    """
    This is the actual state associated with a RunnerEntry.
    It mmust only be managed by the node that actually runs the code.

    Arguments:
      started (float): timestamp when the job was last started
      stopped (float): timestamp when the job last terminated
      result (Any): the code's return value
      node (str): the node running this code
      backoff (float): on error, the multiplier to apply to the restart timeout
    """

    ATTRS = "started stopped result node backoff".split()

    started = 0  # timestamp
    stopped = 0  # timestamp
    node = None  # on which the code is currently running
    result = NotGiven
    backoff = 0

    @property
    def runner(self):
        return self.root.runner.follow(*self.subpath, create=False)

    async def startup(self):
        try:
            run = self.runner
        except KeyError:
            # The code entry doesn't exist any more.
            await self.delete()
            return

        if self.node is None or self.node != self.root.name:
            return

        self.stopped = time.time()
        self.node = None
        self.backoff += 1
        await self.root.runner.err.record_error(
            "run", *self.runner._path, message="Runner restarted"
        )
        await self.save()

    async def stale(self):
        self.stopped = time.time()
        self.node = None
        self.backoff += 2
        await self.root.runner.err.record_error(
            "run",
            *self.runner._path,
            client_name=self.node,
            message="Runner {node} {state}",
            data={"node": self.node, "state": "offline" if self.stopped else "stale"}
        )
        await self.save()

    async def set_value(self, value):
        n = self.node

        await super().set_value(value)
        if not self.root.runner.ready:
            return
        try:
            run = self.runner
        except KeyError:
            return

        # Check whether running code needs to be killed off
        if run.scope is None:
            return
        if self.node is not None and self.node == n:
            # Nothing changed.
            return
        elif self.node is None or n == self.root.runner.name:
            # Owch. Our job got taken away from us.
            self._comment = "Cancel: Node set to %r" % (self.node,)
            await run.scope.cancel()
        elif n is not None:
            logger.warning(
                "Runner %s at %r: running but node is %s",
                self.root.name,
                self.subpath,
                n,
            )

        await run.root.trigger_rescan()


class StateRoot(ClientRoot):
    """Base class for handling the state of entries.

    This is separate from the RunnerRoot hierarchy because the latter may
    be changed by anybody while this subtree may only be affected by the
    actual runner. Otherwise we get interesting race conditions.
    """

    @classmethod
    def child_type(cls, name):
        return StateEntry

    @property
    def name(self):
        return self.client.client_name

    @property
    def runner(self):
        return self._runner()

    def set_runner(self, runner):
        self._runner = ref(runner)

    async def runner_running(self):
        for n in self.all_children:
            await n.startup()

    async def kill_stale_nodes(self, names):
        """States with node names in the "names" set are stale. Kill them."""
        for s in self.all_children:
            if s.node in names:
                await s.stale()

    _last_t = 0
    async def ping(self):
        t = time.time()
        if t-self._last_t >= abs(self._cfg['ping']):
            self._last_t=t
            if self._cfg['ping'] > 0:
                val = self.value_or({},Mapping)
                val['alive'] = t
                await self.update(val)
            else:
                await self.client.msg_send("run", {"group": self.runner.group, "time":t, "node":self.name})


class _BaseRunnerRoot(ClientRoot):
    """common code for AnyRunnerRoot and SingleRunnerRoot"""

    _active: ActorState = None
    _trigger: anyio.abc.Event = None
    _run_now_task: anyio.abc.CancelScope = None

    err = None
    code = None
    ready = False
    _run_next = 0

    CFG = "runner"
    SUB = None

    def __init__(self, *a, _subpath, err=None, code=None, **kw):
        super().__init__(*a, **kw)
        self.err = err
        self.code = code
        self._nodes = {}
        self._trigger = anyio.create_event()
        self._x_subpath = _subpath

    @classmethod
    def child_type(cls, name):
        return RunnerEntry

    @classmethod
    async def as_handler(cls, client, subpath, cfg=None, **kw):
        assert cls.SUB is not None
        if cfg is None:
            cfg_ = client._cfg["runner"]
        else:
            cfg_ = combine_dict(cfg, client._cfg["runner"])
        return await super().as_handler(client, subpath=subpath, _subpath=subpath, cfg=cfg_, **kw)

    async def run_starting(self):
        from .errors import ErrorRoot
        from .code import CodeRoot

        if self.err is None:
            self.err = await ErrorRoot.as_handler(self.client)
        if self.code is None:
            self.code = await CodeRoot.as_handler(self.client)

        await self._state_runner()
        self.state.set_runner(self)

        self.node_history = NodeList(0)
        self._start_delay = self._cfg["start_delay"]

        await super().run_starting()

    async def _state_runner(self):
        self.state = await StateRoot.as_handler(self.client, cfg=self._cfg, subpath=self._x_subpath, key="state")

    @property
    def name(self):
        """my node name"""
        return self.client.client_name

    def get_node(self, name):
        """
        If the runner keeps track of "foreign" nodes, allocate them
        """
        return None

    async def running(self):
        await self._tg.spawn(self._run_actor)

        # the next block needs to be atomic
        for n in self.all_children:
            print(n)
        self.ready = True

        await self.state.runner_running()
        await super().running()

    async def _run_actor(self):
        raise RuntimeError("You want to override me." "")

    async def trigger_rescan(self):
        """Tell the run_now task to rescan our job list"""
        if self._trigger is not None:
            await self._trigger.set()

    async def _run_now(self, evt=None):
        t_next = self._run_next
        async with anyio.open_cancel_scope() as sc:
            self._run_now_task = sc
            if evt is not None:
                await evt.set()
            t = time.time()
            while True:
                if t_next > t:
                    async with anyio.move_on_after(t_next - t):
                        await self._trigger.wait()
                        self._trigger = anyio.create_event()

                t = time.time()
                t_next = t + 99999
                for j in self.all_children:
                    d = j.should_start()
                    if d is False:
                        continue
                    if d <= t:
                        await self._tg.spawn(j.run)
                        await anyio.sleep(self._start_delay)
                    elif t_next > d:
                        t_next = d


class AnyRunnerRoot(_BaseRunnerRoot):
    """
    This class represents the root of a code runner. Its job is to start
    (and periodically restart, if required) the entry points stored under it.
    """

    SUB = "group"

    _stale_times = None

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self.group = "run."+self._name

    def get_node(self, name):
        return RunnerNode(self, name)

    @property
    def max_age(self):
        """Timeout after which we really should have gotten another go"""
        return self._act.cycle_time_max * (self._act.history_size + 1.5)

    async def _run_actor(self):
        async with anyio.create_task_group() as tg:
            self.tg = tg

            async with ClientActor(
                self.client, self.name, topic=self.group, cfg=self._cfg['actor']
            ) as act:
                self._act = act

                self._age_q = anyio.create_queue(10)
                await self.spawn(self._age_killer)

                psutil.cpu_percent(interval=None)
                await act.set_value(0)
                self.seen_load = None

                async for msg in act:
                    logger.debug("Actor %r", msg)
                    if isinstance(msg, PingEvent):
                        await act.set_value(100 - psutil.cpu_percent(interval=None))

                        node = self.get_node(msg.node)
                        node.load = msg.value
                        node.seen = time.time()
                        if self.seen_load is not None:
                            self.seen_load += msg.value
                        self.node_history += node

                    elif isinstance(msg, TagEvent):
                        load = 100 - psutil.cpu_percent(interval=None)
                        await act.set_value(load)
                        if self.seen_load is not None:
                            pass  # TODO

                        self.node_history += self.name
                        evt = anyio.create_event()
                        await self.spawn(self._run_now, evt)
                        await self._age_q.put(None)
                        await evt.wait()

                        await self.state.ping()

                    elif isinstance(msg, UntagEvent):
                        await act.set_value(100 - psutil.cpu_percent(interval=None))
                        self.seen_load = 0

                        await self._run_now_task.cancel()
                        # TODO if this is a DetagEvent, kill everything?
                pass  # end of actor task

    async def find_stale_nodes(self, cur):
        """
        Find stale nodes (i.e. last seen < cur) and clean them.
        """
        if self._stale_times is None:
            self._stale_times = []
        self._stale_times.append(cur)
        if self._stale_times[0] > cur - self.max_age:
            return
        if len(self._stale_times) < 5:
            return
        cut = self._stale_times.pop(0)

        dropped = []
        for node in self._nodes.values():
            if node.seen < cut:
                dropped.append(node)
        names = set()
        for node in dropped:
            names.add(node.name)
            del self._nodes[node.name]
        if names:
            await self.state.kill_stale_nodes(names)

    async def _age_killer(self):
        t1 = time.time()
        while self._age_q is not None:
            await self.find_stale_nodes(t1)
            async with anyio.move_on_after(self.max_age) as sc:
                await self._age_q.get()
                t1 = time.time()
                continue
            t2 = time.time()
            if t1 + self.max_age < t2:
                raise NotSelected(self.max_age, t2, t1)
            t1 = t2

    async def _cleanup_nodes(self):
        t = time.time()
        while len(self.node_history) > 1:
            node = self.get_node(self.node_history[-1])
            if t - node.seen < self.max_age:
                break
            assert node.name == self.node_history.pop()
            for j in self.all_children:
                if j.node == node.name:
                    await j.seems_down()


class SingleRunnerRoot(_BaseRunnerRoot):
    """
    This class represents the root of a code runner. Its job is to start
    (and periodically restart, if required) the entry points stored under it.

    While :class:`AnyRunnerRoot` tries to ensure that the code in question runs
    on any cluster member, this class runs tasks on a single node.
    The code is able to check whether any and/or all of the cluster's main
    nodes are reachable; this way, the code can default to local operation
    if connectivity is lost.

    Local data (dict):

    Arguments:
      cores (tuple): list of nodes whose reachability may determine
        whether the code uses local/emergency/??? mode.
      
    Config file:

    Arguments:
      path (tuple): the location this entry is stored at. Defaults to
        ``('.distkv', 'process')``.
      name (str): this runner's name. Defaults to the client's name plus
        the name stored in the root node, if any.
      actor (dict): the configuration for the underlying actor. See
        ``asyncactor`` for details.
    """

    SUB = "single"

    err = None
    _act = None
    code = None

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self.group = "run."+self._path[-2]

    async def set_value(self, val):
        await super().set_value(val)
        try:
            cores = val["cores"]
        except (TypeError, KeyError):
            if self._act is not None:
                await self._act.disable(0)
        else:
            if self.name in cores:
                await self._act.enable(len(cores))
            else:
                await self._act.disable(len(cores))

    async def notify_active(self):
        """
        Notify all running jobs that there's a change in active status
        """
        oac = self._active
        ac = len(self.node_history)
        if ac == 0:
            ac = DetachedState
        elif self.name in self.node_history and ac == 1:
            ac = DetachedState
        elif ac >= self._act.n_nodes:
            ac = CompleteState
        else:
            ac = PartialState

        if oac is not ac:
            self._active = ac
            for n in self.all_children:
                await n.send_event(ac)

    @property
    def max_age(self):
        """Timeout after which we really should have gotten another ping"""
        return self._act.cycle_time_max * 1.5

    async def _run_actor(self):
        async with anyio.create_task_group() as tg:
            self.tg = tg
            self._age_q = anyio.create_queue(1)

            async with ClientActor(
                self.client, self.name, topic=self.group, cfg=self._cfg
            ) as act:
                self._act = act
                await tg.spawn(self._age_notifier)
                await self.spawn(self._run_now)
                await act.set_value(0)

                async for msg in act:
                    if isinstance(msg, AuthPingEvent):
                        self.node_history += msg.node
                        await self._age_q.put(None)
                        await self.notify_active()

                        await self.state.ping()

                pass  # end of actor task

    async def _age_notifier(self):
        while self._age_q is not None:
            flag = False
            async with anyio.move_on_after(self.max_age):
                await self._age_q.get()
                flag = True
            if not flag:
                await self.notify_active()


class AllRunnerRoot(SingleRunnerRoot):
    """
    This class represents the root of a code runner. Its job is to start
    (and periodically restart, if required) the entry points stored under it.

    This class behaves like `SingleRunner`, except that it runs tasks on all nodes.
    """

    SUB = "all"

    err = None
    _act = None
    code = None

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self.group = "run."+self.name

    async def _state_runner(self):
        self.state = await StateRoot.as_handler(self.client, cfg=self._cfg, subpath=self._x_subpath+(self.client.client_name,), key="state")




import pytest
import trio
import anyio
import time

from .mock_serf import stdtest

from .run import run
from functools import partial

from distkv.client import ServerError
from distkv.util import PathLongener

from distkv.code import CodeRoot
from distkv.runner import AnyRunnerRoot
from distkv.errors import ErrorRoot
import logging

logger = logging.getLogger(__name__)


@pytest.mark.trio
async def test_83_run(autojump_clock):
    from distkv import runner

    async with stdtest(args={"init": 123}) as st:
        s, = st.s
        for h, p, *_ in s.ports:
            if h[0] != ":":
                break
        async with st.client() as c:
            await ErrorRoot.as_handler(c)
            cr = await CodeRoot.as_handler(c)
            r = await AnyRunnerRoot.as_handler(c, subpath=(), code=cr)
            c._test_evt = anyio.create_event()
            await cr.add(
                "forty",
                "two",
                code="""\
                import trio
                c=_client
                await c._test_evt.set()
                await trio.sleep(10)
                return 42
                """,
                is_async=True,
            )
            ru = r.follow("foo", "test", create=True)
            ru.code = ("forty", "two")
            await ru.run_at(time.time()+1)
            logger.info("Start sleep")
            with trio.fail_after(60):
                await c._test_evt.wait()
            await run(
                "-vvv",
                "client",
                "-h",
                h,
                "-p",
                p,
                "data",
                "get",
                "-rd_",
                do_stdout=False,
            )
            await trio.sleep(11)

            logger.info("End sleep")

            r = await run("-vvv", "client", "-h", h, "-p", p, "data", "get")
            assert r.stdout == "123\n"

            rs = ru.state
            assert rs.started > 0
            assert rs.stopped > 0
            assert rs.backoff == 0
            assert rs.result == 42



import pytest
import trio

from .mock_serf import stdtest
from .run import run

# from .run import run
# from functools import partial

from distkv.auth import loader
from distkv.client import ServerError
from distkv.util import PathLongener

from distkv.errors import ErrorRoot
import logging

logger = logging.getLogger(__name__)


async def collect(i, path=()):
    res = []
    pl = PathLongener(path)
    async for r in i:
        r.pop("tock", 0)
        r.pop("seq", 0)
        pl(r)
        res.append(r)
    return res


@pytest.mark.trio
async def test_81_basic(autojump_clock):
    async with stdtest(args={"init": 123}) as st:
        s, = st.s
        async with st.client() as c:
            async with st.client() as cx:
                e = await ErrorRoot.as_handler(c)
                ex = await ErrorRoot.as_handler(cx)
                try:
                    1 / 0
                except Exception as exc:
                    await ex.record_error("tester", "here", "or", "there", exc=exc)
                await trio.sleep(1)
                n = 0
                for err in e.all_errors("tester"):
                    n += 1
                    await err.resolve()
                assert n == 1
                await trio.sleep(1)

@pytest.mark.trio
async def test_82_many(autojump_clock):
    async with stdtest(args={"init": 123}) as st:
        s, = st.s
        async with st.client() as cx, st.client() as cy, st.client() as cz:
                ex = await ErrorRoot.as_handler(cx, name="a1")
                ey = await ErrorRoot.as_handler(cy, name="a2")
                ez = await ErrorRoot.as_handler(cz, name="a3")

                async def err(e):
                    with trio.CancelScope(shield=True):
                        await e.record_error("tester", "dup", message="Owchie at {node}", data={'node':e.name})
                async with trio.open_nursery() as tg:
                    tg.start_soon(err, ex)
                    tg.start_soon(err, ey)
                    tg.start_soon(err, ez)
                    await trio.sleep(2)

                for h, p, *_ in s.ports:
                    if h[0] != ":":
                        break
                r = await run("client", "-m", "-h", h, "-p", p, "data", "get", "-rd_", do_stdout=False)

                n = 0
                for err in ex.all_errors("tester"):
                    n += 1
                    assert len(list(err)) == 3, list(err)
                    for k in err:
                        assert k._name in {"a1","a2","a3"}, k
                    await err.resolve()
                assert n == 1
                await trio.sleep(1)



import pytest
import trio

from .mock_serf import stdtest

# from .run import run
# from functools import partial

from distkv.auth import loader
from distkv.client import ServerError
from distkv.util import PathLongener

import logging

logger = logging.getLogger(__name__)


async def collect(i, path=()):
    res = []
    pl = PathLongener(path)
    async for r in i:
        r.pop("tock", 0)
        r.pop("seq", 0)
        pl(r)
        res.append(r)
    return res


@pytest.mark.trio
async def test_81_basic(autojump_clock):
    async with stdtest(args={"init": 123}) as st:
        s, = st.s
        async with st.client() as c:
            # TODO control what happens when stepping to where's no ACL
            #await c._request("set_internal", path=("acl", "foo"), value="x")
            await c._request("set_internal", path=("acl", "foo", "one"), value="rxnc")
            await c._request(
                "set_internal", path=("acl", "foo", "one", "two"), value="rc"
            )

            um = loader("_test", "user", make=True, server=False)
            u = um.build({"name": "std"})
            await u.send(c)
            u = um.build({"name": "aclix"})
            await u.send(c)
            await c._request("set_internal", path=("auth","_test","user","aclix","acl"), value=dict(key="foo"), iter=False)
            await c._request("set_auth_typ", typ="_test")
            #, "aux": {"acl": "foo"}})


        recv = []
        um = loader("_test", "user", make=False, server=False)

        async with st.client(auth=um.build({"name": "aclix"})) as c:
            await c.set("one", value=10)
            await c.set("one", "two", value=11)
            with pytest.raises(ServerError):
                await c.set("one", "two", "three", value=12)
            with pytest.raises(ServerError):
                await c.set("one", "two", value=22)
#           with pytest.raises(ServerError):
#               await c.set("foo", value=23)



# command line interface

import sys
import asyncclick as click
from distkv.util import MsgReader
from functools import partial
from collections import Mapping

from distkv.util import MsgReader, MsgWriter
from distkv.util import yprint, PathLongener
from distkv.codec import unpacker

import logging

logger = logging.getLogger(__name__)


@main.group(short_help="Manage data.")  # pylint: disable=undefined-variable
@click.pass_obj
async def cli(obj):
    """
    Low-level tools that don't depend on a running server.
    """
    pass


@cli.command()
@click.argument("path", nargs=-1)
@click.pass_obj
async def cfg(obj, path):
    """Emit the current configuration as a YAML file.
    
    You can limit the output by path elements.
    E.g., "cfg connect host" will print "localhost".

    Single values are printed with a trailing line feed.
    """
    cfg = obj.cfg
    for p in path:
        try:
            cfg = cfg[p]
        except KeyError:
            if obj.debug:
                print("Unknown:",p)
            sys.exit(1)
    if isinstance(cfg,str):
        print(cfg, file=obj.stdout)
    else:
        yprint(cfg, stream=obj.stdout)


@cli.command()
@click.option("-p","--path", is_flag=True, default=False,help="Unwrap paths")
@click.argument("file", nargs=1)
@click.pass_obj
async def file(obj, file, path):
    """Read a MsgPack file and dump as YAML."""
    if path:
        pl=PathLongener()
    else:
        pl=lambda _:None
    async with MsgReader(path=file) as f:
        async for msg in f:
            pl(msg)
            yprint(msg, stream=obj.stdout)
            print("---", file=obj.stdout)

@cli.command()
@click.argument("node", nargs=1)
@click.argument("file", type=click.Path(), nargs=1)
@click.pass_obj
async def init(obj, node, file):
    """Write an initial preload file.
    
    Usage: distkv dump init <node> <outfile>

    Writes an initial DistKV file that behaves as if it was generated by <node>.

    Using this command, followed by "distkv server -l <outfile> <node>", is
    equivalent to running "distkv server -i 'Initial data' <node>.
    """
    async with MsgWriter(path=file) as f:
        await f(dict(chain=dict(node=node,tick=1,prev=None),depth=0,path=[],tock=1,value="Initial data"))

@cli.command()
@click.argument("path", nargs=-1)
@click.pass_obj
async def msg(obj, path):
    """
    Monitor the server-to-sever message stream.

    The default is the main server's "update" stream.
    Use '+NAME' to monitor a different stream instead.
    Use '+' to monitor all streams.

    Common streams:
    * ping: sync: all servers (default)
    * update: data changes
    * del: sync: nodes responsible for cleaning up deleted records
    """
    from distkv.backend import get_backend
    import msgpack

    px = 0
    if not path:
        path = obj.cfg.server.root.split('.')
        path.append('update')
    elif len(path) == 1:
        path = path[0].split('.')
        if len(path) == 1 and path[0].startswith('+'):
            p = path[0][1:]
            path = obj.cfg.server.root.split('.') + [p or '#']
            if not p:
                px = len(path)-1
    be = obj.cfg.server.backend
    kw = obj.cfg.server[be]

    async with get_backend(be)(**kw) as conn:
        async with conn.monitor(*path) as stream:
            async for msg in stream:
                v = vars(msg)
                if isinstance(v.get('payload'),(bytearray,bytes)):
                    t = msg.topic
                    v = unpacker(v['payload'])
                    if px > 0:
                        v['_topic'] = t[px:]
                else:
                    v['_type'] = type(msg).__name__
                yprint(v, stream=obj.stdout)
                print("---", file=obj.stdout)



import anyio
import asyncserf
from contextlib import asynccontextmanager
from . import Backend

# Simply setting connect=asyncserf.serf_client interferes with mocking
# when testing.

class SerfBackend(Backend):
    @asynccontextmanager
    async def connect(self, *a, **k):
        async with asyncserf.serf_client(*a,**k) as c:
            self.client = c
            try:
                yield self
            finally:
                self.client = None

    def monitor(self, *topic):
        topic = "user:" + '.'.join(topic)
        return self.client.stream(topic)

    def send(self, *topic, payload):
        """
        Send this payload to this topic.
        """
        return self.client.event('.'.join(topic), payload=payload, coalesce=False)

@asynccontextmanager
async def connect(*a,**kw):
    async with anyio.create_task_group() as tg:
        c = SerfBackend(tg)
        async with c.connect(*a, **kw):
            yield c




# command line interface

import asyncclick as click
import yaml
import sys

from distkv.util import yprint, NotGiven

import logging

logger = logging.getLogger(__name__)


@main.group()  # pylint: disable=undefined-variable
@click.pass_obj
async def cli(obj):
    """Manage code stored in DistKV."""
    pass


@cli.command()
@click.option(
    "-s", "--script", type=click.File(mode="w", lazy=True), help="Save the code here"
)
@click.argument("path", nargs=-1)
@click.pass_obj
async def get(obj, path, script):
    """Read a code entry"""
    if not path:
        raise click.UsageError("You need a non-empty path.")
    res = await obj.client._request(
        action="get_value",
        path=obj.cfg["codes"]["prefix"] + path,
        iter=False,
        nchain=obj.meta,
    )
    if 'value' not in res:
        if obj.debug:
            print("No entry here.", file=sys.stderr)
        sys.exit(1)
    if not obj.meta:
        res = res.value
    if script:
        code = res.pop("code", None)
        if code is not None:
            print(code, file=script)
    yprint(res, stream=obj.stdout)


@cli.command()
@click.option("-a", "--async", "async_", is_flag=True, help="The code is async")
@click.option(
    "-t", "--thread", is_flag=True, help="The code should run in a worker thread"
)
@click.option("-s", "--script", type=click.File(mode="r"), help="File with the code")
@click.option("-v", "--vars", multiple=True, type=str, help="Required variables")
@click.option("-i", "--info", type=str, help="one-liner info about the code")
@click.option(
    "-d", "--data", type=click.File(mode="r"), help="load the metadata (YAML)"
)
@click.argument("path", nargs=-1)
@click.pass_obj
async def set(obj, path, thread, script, data, vars, async_, info):
    """Save Python code."""
    if async_:
        if thread:
            raise click.UsageError("You can't specify both '--async' and '--thread'.")
    else:
        if thread:
            async_ = False
        else:
            async_ = None

    if not path:
        raise click.UsageError("You need a non-empty path.")

    if data:
        msg = yaml.safe_load(data)
    else:
        msg = {}
    chain = NotGiven
    if "value" in msg:
        chain = msg.get("chain", NotGiven)
        msg = msg["value"]
    if async_ is not None or "is_async" not in msg:
        msg["is_async"] = async_

    if info is not None:
        msg["info"] = info

    if "code" in msg:
        if script:
            raise click.UsageError("Duplicate script")
    else:
        if not script:
            raise click.UsageError("Missing script")
        msg["code"] = script.read()

    if "vars" in msg:
        if vars:
            raise click.UsageError("Duplicate variables")
    elif vars:
        vl = msg['vars'] = []
        for vv in vars:
            vl.extend(vv.split(','))

    res = await obj.client.set(
        *obj.cfg["codes"]["prefix"],
        *path,
        value=msg,
        nchain=obj.meta,
        **({'chain':chain} if chain is not NotGiven else {}),
    )
    if obj.meta:
        yprint(res, stream=obj.stdout)


@cli.group("module")
@click.pass_obj
async def mod(obj):
    """
    Change the code of a module stored in DistKV
    """


@mod.command()
@click.option(
    "-s", "--script", type=click.File(mode="w", lazy=True), help="Save the code here"
)
@click.argument("path", nargs=-1)
@click.pass_obj  # pylint: disable=function-redefined
async def get(obj, path, script):
    """Read a module entry"""
    if not path:
        raise click.UsageError("You need a non-empty path.")
    res = await obj.client._request(
        action="get_value",
        path=obj.cfg["modules"]["prefix"] + path,
        iter=False,
        nchain=obj.meta,
    )
    if not obj.meta:
        res = res.value

    code = res.pop("code", None)
    if code is not None:
        code = code.rstrip('\n \t')+"\n"
        if script:
            print(code, file=script)
        else:
            res['code'] = code

    yprint(res, stream=obj.stdout)


@mod.command()
@click.option(
    "-s", "--script", type=click.File(mode="r"), help="File with the module's code"
)
@click.option(
    "-d", "--data", type=click.File(mode="r"), help="load the metadata (YAML)"
)
@click.argument("path", nargs=-1)  # pylint: disable=function-redefined
@click.pass_obj
async def set(obj, path, script, data):
    """Save a Python module to DistKV."""
    if not path:
        raise click.UsageError("You need a non-empty path.")

    if data:
        msg = yaml.safe_load(data)
    else:
        msg = {}
    chain = None
    if "value" in msg:
        chain = msg.get("chain", None)
        msg = msg["value"]

    if "code" not in msg:
        if script:
            raise click.UsageError("Duplicate script")
    else:
        if not script:
            raise click.UsageError("Missing script")
        msg["code"] = script.read()

    res = await obj.client.set(
        *obj.cfg["modules"]["prefix"],
        *path,
        value=msg,
        iter=False,
        nchain=obj.meta,
        chain=chain,
    )
    if obj.meta:
        yprint(res, stream=obj.stdout)


@cli.command()
@click.option(
    "-d",
    "--as-dict",
    default=None,
    help="Structure as dictionary. The argument is the key to use "
    "for values. Default: return as list",
)
@click.option(
    "-m",
    "--maxdepth",
    type=int,
    default=None,
    help="Limit recursion depth. Default: whole tree",
)
@click.option(
    "-M",
    "--mindepth",
    type=int,
    default=None,
    help="Starting depth. Default: whole tree",
)
@click.option("-f", "--full", is_flag=True, help="print complete entries.")
@click.option("-s", "--short", is_flag=True, help="print shortened entries.")
@click.argument("path", nargs=-1)
@click.pass_obj
async def list(obj, path, as_dict, maxdepth, mindepth, full, short):
    """
    List code entries.

    If you read a sub-tree recursively, be aware that the whole subtree
    will be read before anything is printed. Use the "watch --state" subcommand
    for incremental output.
    """

    if (full or as_dict) and short:
        raise click.UsageError("'-f'/'-d' and '-s' are incompatible.")
    kw = {}
    if maxdepth is not None:
        kw["max_depth"] = maxdepth
    if mindepth is not None:
        kw["min_depth"] = mindepth
    y = {}
    async for r in obj.client.get_tree(*obj.cfg['codes'].prefix, *path, nchain=obj.meta, **kw):
        r.pop("seq", None)
        path = r.pop("path")
        if not full:
            if 'info' not in r.value:
                r.value.info = "<%d lines>" % (len(r.value.code.splitlines()),)
            del r.value['code']

        if short:
            print (' '.join(path),"::",r.value.info)
            continue

        if as_dict is not None:
            yy = y
            for p in path:
                yy = yy.setdefault(p, {})
            try:
                yy[as_dict] = r if obj.meta else r.value
            except AttributeError:
                continue
        else:
            y = {}
            try:
                y[path] = r if obj.meta else r.value
            except AttributeError:
                continue
            yprint([y], stream=obj.stdout)

    if as_dict is not None:
        yprint(y, stream=obj.stdout)
    return



# command line interface

import os
import asyncclick as click
import anyio

from distkv.default import CFG
from distkv.server import Server

import logging

logger = logging.getLogger(__name__)


@main.command(short_help="Run the DistKV server.")  # pylint: disable=undefined-variable
@click.option(
    "-l",
    "--load",
    type=click.Path(readable=True, exists=True, allow_dash=False),
    default=None,
    help="Event log to preload.",
)
@click.option(
    "-s",
    "--save",
    type=click.Path(writable=True, allow_dash=False),
    default=None,
    help="Event log to write to.",
    hidden=True,
)
@click.option(
    "-i",
    "--incremental",
    default=None,
    help="Save incremental changes, not the complete state",
    hidden=True,
)
@click.option(
    "-I",
    "--init",
    default=None,
    help="Initial value to set the root to. Use only when setting up "
    "a cluster for the first time!",
    hidden=True,
)
@click.option("-e", "--eval", is_flag=True, help="The 'init' value shall be evaluated.", hidden=True)
@click.argument("name", nargs=1)
@click.pass_obj
async def cli(obj, name, load, save, init, incremental, eval):
    """
    This command starts a DistKV server. It defaults to connecting to the local Serf
    agent.

    All DistKV servers must have a unique name. Its uniqueness cannot be
    verified reliably.

    One server in your network needs either an initial datum, or a copy of
    a previously-saved DistKV state. Otherwise, no client connections will
    be accepted until synchronization with the other servers in your DistKV
    network is complete.

    This command requires a unique NAME argument. The name identifies this
    server on the network. Never start two servers with the same name!
    """

    kw = {}
    if eval:
        kw["init"] = __builtins__["eval"](init)
    elif init == "-":
        kw["init"] = None
    elif init is not None:
        kw["init"] = init

    from distkv.util import as_service

    async with as_service(obj) as evt:
        s = Server(name, cfg=obj.cfg, **kw)
        if load is not None:
            await s.load(path=load, local=True)

        await s.serve(log_path=save, log_inc=incremental, ready_evt=evt)



# command line interface

import os
import sys
import asyncclick as click
import json
from functools import partial

from distkv.util import (
    attrdict,
    combine_dict,
    PathLongener,
    MsgReader,
    PathShortener,
    split_one,
    NotGiven,
)
from distkv.client import open_client
from distkv.command import Loader
from distkv.default import CFG
from distkv.auth import gen_auth

import logging

logger = logging.getLogger(__name__)


class NullObj:
    """
    This helper defers raising an exception until one of its attributes is
    actually accessed.
    """

    def __init__(self, exc):
        self._exc = exc

    def __call__(self, *a, **kw):
        raise self._exc

    def __await__(self):
        raise self._exc

    def __getattr__(self, k):
        if k[0] == "_" and k != "_request":
            return object.__getattribute__(self, k)
        raise self._exc


@main.group(cls=partial(Loader, __file__, "client"))  # pylint: disable=undefined-variable
@click.option(
    "-h", "--host", default=None, help="Host to use. Default: %s" % (CFG.connect.host,)
)
@click.option(
    "-p",
    "--port",
    type=int,
    default=None,
    help="Port to use. Default: %d" % (CFG.connect.port,),
)
@click.option(
    "-a",
    "--auth",
    type=str,
    default=None,
    help="Auth params. =file or 'type param=value…' Default: _anon",
)
@click.option("-m", "--metadata", is_flag=True, help="Include/print metadata.")
@click.pass_context
async def cli(ctx, host, port, auth, metadata):
    """Talk to a DistKV server."""
    obj = ctx.obj
    cfg = attrdict()
    if host is not None:
        cfg.host = host
    if port is not None:
        cfg.port = port

    if auth is not None:
        cfg.auth = gen_auth(auth)
        if obj._DEBUG:
            cfg.auth._DEBUG = True

    cfg = combine_dict(attrdict(connect=cfg), obj.cfg, cls=attrdict)

    obj.meta = 3 if metadata else False

    try:
        obj.client = await ctx.enter_async_context(open_client(**cfg))
    except OSError as exc:
        obj.client = NullObj(exc)
    else:
        logger.debug("Connected.")



"""
This module contains helper functions for packing+unpacking of single messages,
plus an unpacker factory for streams.
"""

import msgpack
from functools import partial

from .util import attrdict

__all__ = ["packer", "unpacker", "stream_unpacker"]

def _encode(data):
    if isinstance(data,int) and data >= 1<<64:
        return msgpack.ExtType(2, data.to_bytes((data.bit_length()+7)//8, 'big'))
    return data

def _decode(code, data):
    if code == 2:
        return int.from_bytes(data, 'big')
    return msgpack.ExtType(code,data)

# single message packer
packer = msgpack.Packer(strict_types=False, use_bin_type=True, default=_encode).pack

# single message unpacker
unpacker = partial(
    msgpack.unpackb, object_pairs_hook=attrdict, raw=False, use_list=False, ext_hook=_decode
)

# stream unpacker factory
stream_unpacker = partial(msgpack.Unpacker,
    object_pairs_hook=attrdict, raw=False, use_list=False, ext_hook=_decode
)



# command line interface

import sys
import asyncclick as click
import time
import anyio

from distkv.exceptions import ServerError
from distkv.code import CodeRoot
from distkv.runner import AnyRunnerRoot, SingleRunnerRoot, AllRunnerRoot
from distkv.util import yprint, PathLongener

import logging

logger = logging.getLogger(__name__)


@main.group()  # pylint: disable=undefined-variable
@click.option("-n", "--node", help="node to run this code on. Empty: any one node, '-': all nodes")
@click.option("-g", "--group", help="group to run this code on. Empty: any one node")
@click.pass_obj
async def cli(obj, node, group):
    """Run code stored in DistKV."""
    if group is None:
        group = "default"
    if not node:
        obj.runner_root = AnyRunnerRoot
        subpath = (group,)
    elif node == '-':
        obj.runner_root = AllRunnerRoot
        subpath = (group,)
    else:
        obj.runner_root = SingleRunnerRoot
        subpath = (node, group)

    obj.subpath = (obj.cfg["runner"]["sub"][obj.runner_root.SUB],) + subpath
    obj.path = obj.cfg["runner"]["prefix"] + obj.subpath
    obj.statepath = obj.cfg["runner"]["state"] + obj.subpath


@cli.command()
@click.pass_obj
async def all(obj):
    """
    Run code that needs to run.

    This does not return.
    """
    from distkv.util import as_service
    if obj.subpath[-1] == "-":
        raise click.UsageError("Group '-' can only be used for listing.")

    async with as_service(obj) as evt:
        _, evt = evt
        c = obj.client
        cr = await CodeRoot.as_handler(c)
        r = await obj.runner_root.as_handler(c, subpath=obj.subpath, code=cr)
        await evt.set()
        while True:
            await anyio.sleep(99999)


@cli.command()
@click.option("-s", "--state", is_flag=True, help="Add state data")
@click.option(
    "-d",
    "--as-dict",
    default=None,
    help="Structure as dictionary. The argument is the key to use "
    "for values. Default: return as list",
)
@click.argument("path", nargs=-1)
@click.pass_obj
async def list(obj, state, as_dict, path):
    """List run entries.
    """
    if obj.subpath[-1] == "-":
        if path:
            raise click.UsageError("Group '-' can only be used without a path.")

        path = obj.path[:-1]
        res = await obj.client._request(
            action="get_tree", path=path, iter=True, max_depth=2,nchain=obj.meta, empty=True
        )
        pl=PathLongener(())
        async for r in res:
            pl(r)
            print(r.path[-1],file=obj.stdout)
        return

    if not path:
        path = ()
    path = obj.path+path
    if state:
        state = obj.statepath+path
    res = await obj.client._request(
        action="get_tree", path=path, iter=True, nchain=obj.meta
    )

    y = {}
    async for r in res:
        if as_dict is not None:
            yy = y
            for p in r.pop("path"):
                yy = yy.setdefault(p, {})
            yy[as_dict] = r if obj.meta else r.pop("value")
        else:
            yy = {}
            if obj.meta:
                yy[r.pop("path")] = r
            else:
                yy[r.path] = r.value

        if state:
            rs = await obj.client._request(
                action="get_value", path=state, iter=False, nchain=obj.meta
            )
            if "value" in rs:
                if not obj.meta:
                    rs = rs.value
                yy["state"] = rs
            else:
                yy["state"] = None
        if as_dict is None:
            yprint([yy], stream=obj.stdout)

    if as_dict is not None:
        yprint(y, stream=obj.stdout)


@cli.command()
@click.option("-r", "--result", is_flag=True, help="Just print the actual result.")
@click.argument("path", nargs=-1)
@click.pass_obj
async def state(obj, path, result):
    """Get the status of a runner entry.
    """
    if obj.subpath[-1] == "-":
        raise click.UsageError("Group '-' can only be used for listing.")
    if result and obj.meta:
        raise click.UsageError("You can't use '-v' and '-r' at the same time.")
    if not path:
        raise click.UsageError("You need a non-empty path.")
    path = obj.statepath + path

    res = await obj.client._request(
        action="get_value", path=path, iter=False, nchain=obj.meta
    )
    if "value" not in res:
        if obj.debug:
            print("Not found (yet?)", file=sys.stderr)
        sys.exit(1)
    if not obj.meta:
        res = res.value

    yprint(res, stream=obj.stdout)


@cli.command()
@click.argument("path", nargs=-1)
@click.pass_obj
async def get(obj, path):
    """Read a runner entry"""
    if obj.subpath[-1] == "-":
        raise click.UsageError("Group '-' can only be used for listing.")
    if not path:
        raise click.UsageError("You need a non-empty path.")
    path = obj.path+ path

    res = await obj.client._request(
        action="get_value", path=path, iter=False, nchain=obj.meta
    )
    if not obj.meta:
        res = res.value

    yprint(res, stream=obj.stdout)


@cli.command()
@click.option(
    "-c", "--code", help="Path to the code that should run. Space separated path."
)
@click.option("-t", "--time", "tm", type=float, help="time the code should next run at")
@click.option("-r", "--repeat", type=int, help="Seconds the code should re-run after")
@click.option("-k", "--ok", type=int, help="Code is OK if it ran this many seconds")
@click.option("-b", "--backoff", type=float, help="Back-off factor. Default: 1.4")
@click.option(
    "-d", "--delay", type=int, help="Seconds the code should retry after (w/ backoff)"
)
@click.option("-i", "--info", help="Short human-readable information")
@click.option(
    "-e", "--eval", "eval_", help="'code' is a Python expression (must eval to a list)"
)
@click.argument("path", nargs=-1)
@click.pass_obj
async def set(obj, path, code, eval_, tm, info, ok, repeat, delay, backoff):
    """Save / modify a run entry."""
    if obj.subpath[-1] == "-":
        raise click.UsageError("Group '-' can only be used for listing.")
    if not path:
        raise click.UsageError("You need a non-empty path.")
    if eval_:
        code = eval(code)
        if not isinstance(code, (list, tuple)):
            raise click.UsageError("'code' must be a list")
    elif code is not None:
        code = code.split(" ")

    path = obj.path + path

    try:
        res = await obj.client._request(
            action="get_value", path=path, iter=False, nchain=3
        )
        if "value" not in res:
            raise ServerError
    except ServerError:
        if code is None:
            raise click.UsageError("New entry, need code")
        res = {}
        chain = None
    else:
        chain = res["chain"]
        res = res["value"]

    if code is not None:
        res["code"] = code
    if ok is not None:
        res["ok_after"] = ok
    if info is not None:
        res["info"] = info
    if backoff is not None:
        res["backoff"] = backoff
    if delay is not None:
        res["delay"] = delay
    if repeat is not None:
        res["repeat"] = repeat
    if tm is not None:
        res["target"] = time.time() + tm

    res = await obj.client.set(
        *path,
        value=res,
        nchain=3,
        **({"chain": chain} if obj.meta else {})
    )
    if obj.meta:
        yprint(res, stream=obj.stdout)

@click.command()
@click.pass_obj
async def monitor(obj):
    """
    Runners periodically send a keepalive message to Serf, if thus
    configured (this is the default).
    """

    # TODO this does not watch changes in DistKV.
    # It also should watch individual jobs' state changes.
    if obj.subpath[-1] == "-":
        raise click.UsageError("Group '-' can only be used for listing.")

    async with obj.client.msg_monitor("run") as cl:
        async for msg in cl:
            yprint(msg, stream=obj.stdout)
            print("---", file=obj.stdout)
            await process_test(msg.data)




import pytest
import trio
import time

from .mock_serf import stdtest

from .run import run
from functools import partial

from distkv.client import ServerError
from distkv.util import PathLongener

from distkv.code import CodeRoot, ModuleRoot
from distkv.errors import ErrorRoot
import logging

logger = logging.getLogger(__name__)


@pytest.mark.trio
async def test_81_basic(autojump_clock):
    async with stdtest(args={"init": 123}) as st:
        s, = st.s
        async with st.client() as c:
            await ErrorRoot.as_handler(c)
            cr = await CodeRoot.as_handler(c)
            await cr.add("forty", "two", code="return 42")
            assert cr("forty.two") == 42


@pytest.mark.trio
async def test_82_module(autojump_clock):
    async with stdtest(args={"init": 123}) as st:
        s, = st.s
        async with st.client() as c:
            await ErrorRoot.as_handler(c)
            m = await ModuleRoot.as_handler(c)
            await m.add(
                "bar",
                "baz",
                code="""
def quux():
    return 42
""",
            )
            cr = await CodeRoot.as_handler(c)
            await cr.add(
                "forty",
                "two",
                code="""
from bar.baz import quux
return quux()
""",
            )
            assert cr("forty.two") == 42



"""
This module affords all DistKV exceptions.
"""

# pylint: disable=unnecessary-pass


class DistKVError(RuntimeError):
    """Superclass of all DistKV errors.

    Abstract class.
    """

    pass


class ServerError(DistKVError):
    """Generic server error.

    This class includes errors forwarded to the client.
    """

    pass


class ClientError(DistKVError):
    """Generic client error.

    Abstract class.
    """

    pass


class ClientConnectionError(ClientError):
    """Some connection error"""

    pass


class ServerClosedError(ServerError):
    """The server closed our connection."""

    pass


class ServerConnectionError(ServerError):
    """Some connection error"""

    pass


class ACLError(ServerError):
    """An ACL did not match"""

    pass


class CancelledError(ClientError):
    """A client call was cancelled."""

    pass


class ClientAuthError(ClientError):
    """Authorization failed.

    Abstract class.
    """

    pass


class ClientAuthRequiredError(ClientAuthError):
    """Authorization required but missing."""

    pass


class ClientAuthMethodError(ClientAuthError):
    """Wrong authorization method provided."""

    pass


class DistKVauthError(ClientError):
    """Auth error.

    Abstract class.
    """

    pass


class NoAuthError(DistKVauthError):
    """Server-side error: auth required"""

    pass


class NoAuthModuleError(DistKVauthError):
    """Server-side error: auth module doesn't exist"""

    pass


class AuthFailedError(DistKVauthError):
    """Server-side error: auth failed"""

    pass



import pytest
import trio

from .mock_serf import stdtest
from .run import run

# from .run import run
# from functools import partial

from distkv.auth import loader
from distkv.client import ServerError
from distkv.util import PathLongener

import logging

logger = logging.getLogger(__name__)


@pytest.mark.trio
async def test_81_basic(autojump_clock):
    async with stdtest(args={"init": 123}) as st:
        s, = st.s
        async with st.client() as c:
            assert 'hoo' not in c.config
            res = await c.set(".distkv","config","hoo", value={"hello":"there"}, nchain=2)
            await c._config.wait_chain(res.chain)
            assert c.config.hoo['hello'] == "there"

        # TODO test iterating on changes
